@@ -11,7 +11,6 @@ from typing import (
     Mapping,
     MutableMapping,
     NamedTuple,
-    Optional,
     Tuple,
     Union,
 )
@@ -20,8 +19,8 @@ from mdformat import codepoints
 from mdformat.renderer._util import (
     CONSECUTIVE_KEY,
     RE_CHAR_REFERENCE,
-    decimalify_leading_whitespace,
-    decimalify_trailing_whitespace,
+    decimalify_leading,
+    decimalify_trailing,
     escape_asterisk_emphasis,
     escape_underscore_emphasis,
     get_list_marker_type,
@@ -86,12 +85,22 @@ def html_inline(node: "RenderTreeNode", context: "RenderContext") -> str:
     return node.content
 
 
+def _in_block(block_name: str, node: "RenderTreeNode") -> bool:
+    while node.parent:
+        if node.parent.type == block_name:
+            return True
+        node = node.parent
+    return False
+
+
 def hardbreak(node: "RenderTreeNode", context: "RenderContext") -> str:
+    if _in_block("heading", node):
+        return "<br /> "
     return "\\" + "\n"
 
 
 def softbreak(node: "RenderTreeNode", context: "RenderContext") -> str:
-    if context.do_wrap and _in_paragraph(node):
+    if context.do_wrap and _in_block("paragraph", node):
         return WRAP_POINT
     return "\n"
 
@@ -119,9 +128,7 @@ def text(node: "RenderTreeNode", context: "RenderContext") -> str:
 
     # Escape "&" if it starts a sequence that can be interpreted as
     # a character reference.
-    for char_refs_found, char_ref in enumerate(RE_CHAR_REFERENCE.finditer(text)):
-        start = char_ref.start() + char_refs_found
-        text = text[:start] + "\\" + text[start:]
+    text = RE_CHAR_REFERENCE.sub(r"\\\1", text)
 
     # The parser can give us consecutive newlines which can break
     # the markdown structure. Replace two or more consecutive newlines
@@ -134,7 +141,7 @@ def text(node: "RenderTreeNode", context: "RenderContext") -> str:
     if text.endswith("!") and next_sibling and next_sibling.type == "link":
         text = text[:-1] + "\\!"
 
-    if context.do_wrap and _in_paragraph(node):
+    if context.do_wrap and _in_block("paragraph", node):
         text = re.sub(r"\s+", WRAP_POINT, text)
 
     return text
@@ -142,7 +149,7 @@ def text(node: "RenderTreeNode", context: "RenderContext") -> str:
 
 def fence(node: "RenderTreeNode", context: "RenderContext") -> str:
     info_str = node.info.strip()
-    lang = info_str.split()[0] if info_str.split() else ""
+    lang = info_str.split(maxsplit=1)[0] if info_str else ""
     code_block = node.content
 
     # Info strings of backtick code fences can not contain backticks or tildes.
@@ -160,9 +167,7 @@ def fence(node: "RenderTreeNode", context: "RenderContext") -> str:
         except Exception:
             # Swallow exceptions so that formatter errors (e.g. due to
             # invalid code) do not crash mdformat.
-            assert (
-                node.map is not None
-            ), "A fence token must always have `map` attribute set"
+            assert node.map is not None, "A fence token must have `map` attribute set"
             LOGGER.warning(
                 f"Failed formatting content of a {lang} code block "
                 f"(line {node.map[0] + 1} before formatting)"
@@ -183,15 +188,20 @@ def code_block(node: "RenderTreeNode", context: "RenderContext") -> str:
 def image(node: "RenderTreeNode", context: "RenderContext") -> str:
     description = _render_inline_as_text(node, context)
 
+    if context.do_wrap:
+        # Prevent line breaks
+        description = description.replace(WRAP_POINT, " ")
+
     ref_label = node.meta.get("label")
     if ref_label:
-        context.env.setdefault("used_refs", set()).add(ref_label)
+        context.env["used_refs"].add(ref_label)
         ref_label_repr = ref_label.lower()
         if description.lower() == ref_label_repr:
             return f"![{description}]"
         return f"![{description}][{ref_label_repr}]"
 
     uri = node.attrs["src"]
+    assert isinstance(uri, str)
     uri = maybe_add_link_brackets(uri)
     title = node.attrs.get("title")
     if title is not None:
@@ -218,6 +228,7 @@ def _render_inline_as_text(node: "RenderTreeNode", context: "RenderContext") ->
             "text": text_renderer,
             "image": image_renderer,
             "link": link,
+            "softbreak": softbreak,
         },
     )
     inline_context = RenderContext(
@@ -238,17 +249,19 @@ def link(node: "RenderTreeNode", context: "RenderContext") -> str:
 
     ref_label = node.meta.get("label")
     if ref_label:
-        context.env.setdefault("used_refs", set()).add(ref_label)
+        context.env["used_refs"].add(ref_label)
         ref_label_repr = ref_label.lower()
         if text.lower() == ref_label_repr:
             return f"[{text}]"
         return f"[{text}][{ref_label_repr}]"
 
     uri = node.attrs["href"]
+    assert isinstance(uri, str)
     uri = maybe_add_link_brackets(uri)
     title = node.attrs.get("title")
     if title is None:
         return f"[{text}]({uri})"
+    assert isinstance(title, str)
     title = title.replace('"', '\\"')
     return f'[{text}]({uri} "{title}")'
 
@@ -277,7 +290,7 @@ def heading(node: "RenderTreeNode", context: "RenderContext") -> str:
 
     # There can be newlines in setext headers, but we make an ATX
     # header always. Convert newlines to spaces.
-    text = text.replace("\n", " ").rstrip()
+    text = text.replace("\n", " ")
 
     # If the text ends in a sequence of hashes (#), the hashes will be
     # interpreted as an optional closing sequence of the heading, and
@@ -289,31 +302,20 @@ def heading(node: "RenderTreeNode", context: "RenderContext") -> str:
 
 
 def blockquote(node: "RenderTreeNode", context: "RenderContext") -> str:
-    text = make_render_children(separator="\n\n")(node, context)
-    lines = text.splitlines()
-    if not lines:
-        return ">"
-    quoted_lines = (f"> {line}" if line else ">" for line in lines)
-    quoted_str = "\n".join(quoted_lines)
-    return quoted_str
+    marker = "> "
+    indent_width = len(marker)
+    context.env["indent_width"] += indent_width
 
-
-def _first_line_width(text: str) -> int:
-    width = 0
-    for c in text:
-        if c == "\n":
-            return width
-        width += 1
-    return width
-
-
-def _last_line_width(text: str) -> int:
-    width = 0
-    for c in reversed(text):
-        if c == "\n":
-            return width
-        width += 1
-    return width
+    try:
+        text = make_render_children(separator="\n\n")(node, context)
+        lines = text.splitlines()
+        if not lines:
+            return ">"
+        quoted_lines = (f"{marker}{line}" if line else ">" for line in lines)
+        quoted_str = "\n".join(quoted_lines)
+        return quoted_str
+    finally:
+        context.env["indent_width"] -= indent_width
 
 
 def _wrap(text: str, *, width: Union[int, Literal["no"]]) -> str:
@@ -368,30 +370,28 @@ def _recover_preserve_chars(text: str, replacements: str) -> str:
     )
 
 
-def _in_paragraph(node: "RenderTreeNode") -> bool:
-    while node.parent:
-        if node.parent.type == "paragraph":
-            return True
-        node = node.parent
-    return False
-
-
 def paragraph(node: "RenderTreeNode", context: "RenderContext") -> str:  # noqa: C901
     inline_node = node.children[0]
+    text = inline_node.render(context)
 
-    wrap_mode = context.options.get("mdformat", {}).get("wrap", "keep")
-    if isinstance(wrap_mode, int) or wrap_mode == "no":
-        text = "".join(child.render(context) for child in inline_node.children)
+    if context.do_wrap:
+        wrap_mode = context.options["mdformat"]["wrap"]
+        if isinstance(wrap_mode, int):
+            wrap_mode -= context.env["indent_width"]
+            wrap_mode = max(1, wrap_mode)
         text = _wrap(text, width=wrap_mode)
-    else:
-        text = inline_node.render(context)
+
+    # A paragraph can start or end in whitespace e.g. if the whitespace was
+    # in decimal representation form. We need to re-decimalify it, one reason being
+    # to enable "empty" paragraphs with whitespace only.
+    text = decimalify_leading(codepoints.UNICODE_WHITESPACE, text)
+    text = decimalify_trailing(codepoints.UNICODE_WHITESPACE, text)
 
     lines = text.split("\n")
     for i in range(len(lines)):
-        # Replace line starting tabs with numeric decimal representation.
-        # A normal tab character would start a code block.
-        if lines[i].startswith("\t"):
-            lines[i] = "&#9;" + lines[i][1:]
+        # Strip whitespace to prevent issues like a line starting tab that is
+        # interpreted as start of a code block.
+        lines[i] = lines[i].strip()
 
         # If a line looks like an ATX heading, escape the first hash.
         if re.match(r"#{1,6}( |\t|$)", lines[i]):
@@ -437,9 +437,6 @@ def paragraph(node: "RenderTreeNode", context: "RenderContext") -> str:  # noqa:
 
     text = "\n".join(lines)
 
-    text = decimalify_leading_whitespace(text)
-    text = decimalify_trailing_whitespace(text)
-
     return text
 
 
@@ -449,8 +446,7 @@ def list_item(node: "RenderTreeNode", context: "RenderContext") -> str:
     This returns just the content. List item markers and indentation are
     added in `bullet_list` and `ordered_list` renderers.
     """
-    is_tight = is_tight_list_item(node)
-    block_separator = "\n" if is_tight else "\n\n"
+    block_separator = "\n" if is_tight_list_item(node) else "\n\n"
     text = make_render_children(block_separator)(node, context)
 
     if not text.strip():
@@ -463,6 +459,9 @@ def bullet_list(node: "RenderTreeNode", context: "RenderContext") -> str:
     first_line_indent = " "
     indent = " " * len(marker_type + first_line_indent)
     block_separator = "\n" if is_tight_list(node) else "\n\n"
+
+    context.env["indent_width"] += len(indent)
+
     text = ""
     for child_idx, child in enumerate(node.children):
         list_item = child.render(context)
@@ -480,6 +479,8 @@ def bullet_list(node: "RenderTreeNode", context: "RenderContext") -> str:
         text += "\n".join(formatted_lines)
         if child_idx != len(node.children) - 1:
             text += block_separator
+
+    context.env["indent_width"] -= len(indent)
     return text
 
 
@@ -490,9 +491,18 @@ def ordered_list(node: "RenderTreeNode", context: "RenderContext") -> str:
     block_separator = "\n" if is_tight_list(node) else "\n\n"
     list_len = len(node.children)
 
-    starting_number: Optional[int] = node.attrs.get("start")
+    starting_number = node.attrs.get("start")
     if starting_number is None:
         starting_number = 1
+    assert isinstance(starting_number, int)
+
+    if consecutive_numbering:
+        indent_width = len(
+            f"{list_len + starting_number - 1}{marker_type}{first_line_indent}"
+        )
+    else:
+        indent_width = len(f"{starting_number}{marker_type}{first_line_indent}")
+    context.env["indent_width"] += indent_width
 
     text = ""
     for list_item_index, list_item in enumerate(node.children):
@@ -510,7 +520,6 @@ def ordered_list(node: "RenderTreeNode", context: "RenderContext") -> str:
             #   112. Last item
             number = starting_number + list_item_index
             pad = len(str(list_len + starting_number - 1))
-            indentation = " " * (pad + len(f"{marker_type}{first_line_indent}"))
             number_str = str(number).rjust(pad, "0")
             formatted_lines.append(
                 f"{number_str}{marker_type}{first_line_indent}{first_line}"
@@ -530,7 +539,6 @@ def ordered_list(node: "RenderTreeNode", context: "RenderContext") -> str:
             other_item_marker = (
                 "0" * (len(str(starting_number)) - 1) + "1" + marker_type
             )
-            indentation = " " * len(first_item_marker + first_line_indent)
             if list_item_index == 0:
                 formatted_lines.append(
                     f"{first_item_marker}{first_line_indent}{first_line}"
@@ -544,11 +552,13 @@ def ordered_list(node: "RenderTreeNode", context: "RenderContext") -> str:
                     else other_item_marker
                 )
         for line in line_iterator:
-            formatted_lines.append(f"{indentation}{line}" if line else "")
+            formatted_lines.append(" " * indent_width + line if line else "")
 
         text += "\n".join(formatted_lines)
         if list_item_index != len(node.children) - 1:
             text += block_separator
+
+    context.env["indent_width"] -= indent_width
     return text
 
 
